MON            (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)
DAY            ([ 0-3][0-9])
TIME           ([ 0-2][0-9]\:[0-6][0-9]\:[0-6][0-9])
TS             {MON}" "{DAY}" "{TIME}
TZ             " "+\+[^ \n]{1,4}
O              ([0-9]|[0-9][0-9]|[0-1][0-9][0-9]|2[0-4][0-9]|25[0-5])
IP             {O}((\.{O}){3})
LOCALHOST      (127\.0\.0\.1)
FIRST          [A-Za-z0-9]
MIDDLE         [-.A-Za-z0-9]
LAST           [A-Za-z0-9]
HN             ({FIRST}{MIDDLE}*{LAST}|{FIRST}{MIDDLE}*|{MIDDLE}*{LAST})
NSP            [^ \n]

    #include <sys/types.h>
    #include <sys/socket.h>
    #include <sys/stat.h>

    #include <netinet/in.h>

    #include <arpa/inet.h>
    #include <arpa/nameser.h>

    #include <ctype.h>
    #include <errno.h>
    #include <netdb.h>
    #include <stdio.h>
    #include <string.h>
    #include <time.h>
    #include <unistd.h>

    #ifdef notdef
    #ifdef HAVE_OS_PROTO_H
    #include "os-proto.h"
    #endif
    #endif

    #undef yywrap
    #define YY_NO_INPUT

    int yywrap(void);
    int yylex(void);
    void addr(const char *, size_t);
    void addr1(const char *);
    void name(const char *, size_t);
    void summary(void);

    const char *p;
    int i;
    int debug;
    int doseek;
    int doinode;
    int timestamp;
    int stats;
    ino_t inode;
    off_t offset;
    ino_t lastinode;
    off_t lastoffset;
    char junkchar = '\0';
    time_t lastts;
    u_int lineno = 1;
    const char *file = "?";

%option nounput
%x ident eat junk

%%

<INITIAL>^{TS}{TZ}" "	BEGIN(ident);
<INITIAL>^{TS}" "	BEGIN(ident);

	/* Lines that don't start with a timestamp are junk */

<INITIAL>^.		{
			junkchar = *yytext;
			BEGIN(junk);
			}

<junk>.*		{
			if (debug)
				fprintf(stderr, "%s:%u JUNK: \"%c%s\"\n",
				    file, lineno, junkchar, yytext);
			}

	/* There are many ways to pick up the syslog host IP */

<ident>{IP}" "		{
			addr(yytext, yyleng - 1);
			BEGIN(eat);
			}

<ident>{HN}" "		{
			name(yytext, yyleng - 1);
			BEGIN(eat);
			}

<ident>{IP}"/"{LOCALHOST}"/"{IP}" "	{
			/* Save 1st IP in chain */
			addr1(yytext);
			BEGIN(eat);
			}

<ident>{IP}"/"{IP}"/"{IP}" "	{
			/* Save 2nd IP in chain */
			p = strchr(yytext, '/');
			if (p == NULL)
				abort();
			++p;
			addr1(p);
			BEGIN(eat);
			}

<ident>{HN}"/"{LOCALHOST}"/"{IP}" "	{
			/* Save 1st hostname in chain */
			p = strchr(yytext, '/');
			if (p == NULL)
				abort();
			i = p - yytext;
			name(yytext, i);
			BEGIN(eat);
			}

<ident>{HN}"/"{IP}"/"{IP}" "	{
			/* Save 2nd IP in chain */
			p = strchr(yytext, '/');
			if (p == NULL)
				abort();
			++p;
			addr1(p);
			BEGIN(eat);
			}

<ident>{IP}"/"{LOCALHOST}" " {
			/* Save 1st ip address in chain */
			p = strchr(yytext, '/');
			if (p == NULL)
				abort();
			i = p - yytext;
			addr(yytext, i);
			BEGIN(eat);
			}

<ident>{LOCALHOST}"/"{IP}" " {
			/* Grab 2nd IP in chain */
			p = strchr(yytext, '/');
			if (p == NULL)
				abort();
			++p;
			i = yyleng - (p - yytext);
			addr(p, i - 1);
			BEGIN(eat);
			}

<ident>{IP}"/"{IP}" "	{
			/* Grab 2nd IP in chain */
			p = strchr(yytext, '/');
			if (p == NULL)
				abort();
			++p;
			i = yyleng - (p - yytext);
			addr(p, i - 1);
			BEGIN(eat);
			}

<ident>{HN}"/"{LOCALHOST}" " {
			/* Save 1st hostname in chain */
			p = strchr(yytext, '/');
			if (p == NULL)
				abort();
			i = p - yytext;
			name(yytext, i);
			BEGIN(eat);
			}

<ident>{HN}"/"{IP}" "	{
			/* Grab 2nd IP in chain */
			p = strchr(yytext, '/');
			if (p == NULL)
				abort();
			++p;
			i = yyleng - (p - yytext);
			addr(p, i - 1);
			BEGIN(eat);
			}

	/* /Applications/Retrospect/131.243.10.39 6.1/Retrospect[...] */
	/* /System/Library/CoreServices/ARD/131.243.10.53 Agent.app[...] */

<ident>{NSP}+"/"{IP}" " {
			/* Just grab the last IP */
			p = strrchr(yytext, '/');
			if (p == NULL)
				abort();
			++p;
			i = yyleng - (p - yytext);
			/* Don't save trailing blank */
			addr(p, i - 1);
			BEGIN(eat);
			}

<ident>s_internal@syslog" "	BEGIN(eat);

<ident>"syslog/syslog "	BEGIN(eat);

<*>\n			{
			++lineno;
			junkchar = '\0';
			BEGIN(INITIAL);
			}

<eat>[^\n]*		/* eat */

<*>.			{
			/* XXX shouldn't see many of these */
			if (debug)
				fprintf(stderr, "{%s:%u %s}\n",
				    file, lineno, yytext);
			}

%%

/*
 * Copyright (c) 2007, 2008, 2009, 2010, 2015, 2016, 2017, 2020
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char copyright[] __attribute__ ((unused)) =
    "@(#) Copyright (c) 2007, 2008, 2009, 2010, 2015, 2016, 2017, 2020\n\
The Regents of the University of California.  All rights reserved.\n";
static const char rcsid[] __attribute__ ((unused)) =
    "@(#) $Id$ (LBL)";
#endif

char *prog;

int errors;

int targc;
char **targv;

extern char *optarg;
extern int optind, opterr;

#define HSIZE 8192			/* must be a power of two */

/* Address hash table */
struct htable {
	struct htable *next;
	char *name;
	int count;
	int state;
	u_int32_t addr;
} htable[HSIZE];

/* Simple name hash table (points back to address hash table) */
struct htable2 {
	struct htable2 *next;
	char *host;
	struct htable *htablep;
	int state;
} htable2[HSIZE];

#define STATE_FREE	0	/* not in use */
#define STATE_USED	1

/* Forwards */
int main(int, char **);
int compipv4(const void *, const void *);
struct htable *cacheaddr(u_int32_t, int, const char *);
struct htable2 *cachehost(const char *, int);
static inline struct htable2 *hash2(const char *);
static inline struct htable *hash(u_int32_t);
struct htable *increment(u_int32_t, const char *);
void logseek(FILE *, const char *, off_t *);
struct htable *lookupaddr(u_int32_t);
struct htable2 *lookuphost(const char *);
void lowcase(char *);
void saveinfo(FILE *);
void usage(void) __attribute__((noreturn));

/* Globals */
size_t numlst;
size_t sizelst;
struct in_addr *srclst = NULL;
struct in_addr *dstlst = NULL;
int addrcnt;				/* number of gethostbyaddr() calls */
int hostcnt;				/* number of gethostbyname() calls */
int totalhosts;

int
main(argc, argv)
	int argc;
	char **argv;
{
	char *cp, *ep;
	int op;
	long ll;

	if (argv[0] == NULL)
		prog = "hf";
	else if ((cp = strrchr(argv[0], '/')) != NULL)
		prog = cp + 1;
	else
		prog = argv[0];

	opterr = 0;
	while ((op = getopt(argc, argv, "dtsI:O:")) != EOF)
		switch (op) {

		case 'd':
			++debug;
			break;

		case 't':
			++timestamp;
			break;

		case 's':
			++stats;
			break;

		case 'I':
			ll = strtoll(optarg, &ep, 10);
			if (*ep != '\0')
				usage();
			inode = ll;
			if (inode != ll)
				usage();
			++doinode;
			break;

		case 'O':
			ll = strtoll(optarg, &ep, 10);
			if (*ep != '\0')
				usage();
			offset = ll;
			if (offset != ll)
				usage();
			++doseek;
			break;

		default:
			usage();
		}

	/* Let yywrap() figure out if there are any arguments to open */
	targc = argc - optind;
	targv = &argv[optind];
	yyin = NULL;
	(void)yywrap();

	/* Process file opened by yywrap() or stdin if no arguments */
	if (yyin)
		yylex();

	if (!errors)
		summary();
	exit(errors);
}

int
yywrap()
{
	static int didany = 0;

	/* Close file, if necessary */
	if (yyin) {
		saveinfo(yyin);
		if (yyin != stdin)
			(void)fclose(yyin);
		yyin = NULL;
	}

	/* Spin through arguments until we run out or successfully open one */
	while (targc > 0) {
		file = targv[0];
		--targc;
		++targv;
		++didany;
		if ((yyin = fopen(file, "r")) != NULL) {
			logseek(yyin, file, &offset);
			return (0);
		}
		perror(file);
		++errors;
	}
	if (!didany) {
		++didany;
		yyin = stdin;
		file = "<stdin>";
		logseek(yyin, file, &offset);
		return (0);
	}
	return (1);
}

/* "IP" with possible trailing junk */
void
addr(const char *str, size_t len)
{
	struct in_addr a;
	char taddr[32];

	if (len > sizeof(taddr) - 1)
		len = sizeof(taddr) - 1;
	strncpy(taddr, str, len);
	taddr[len] = '\0';
	if (inet_aton(taddr, &a) == 0)
		abort();
	(void)increment(a.s_addr, NULL);
}

/* "IP/IP" with possible trailing junk; take first IP */
void
addr1(const char *str)
{
	const char *p;
	size_t len;

	p = strchr(str, '/');
	if (p == NULL)
		abort();
	len = p - str;
	addr(str, len);
}

struct htable *
cacheaddr(u_int32_t a, int state, const char *fqdn)
{
	struct htable *p, *p2;

	/* Don't cache zero */
	if (a == 0)
		return (NULL);

	/* Look for existing slot in hash table */
	for (p = hash(a); p != NULL; p = p->next)
		if (p->state != STATE_FREE && a == p->addr)
			break;

	/* Allocate a new slot */
	if (p == NULL) {
		p = hash(a);
		if (p->state != STATE_FREE) {
			/* Handle the collision */
			p2 = (struct htable *)malloc(sizeof(*p2));
			/* Lose, lose */
			if (p2 == NULL)
				return (NULL);
			memset((char *)p2, 0, sizeof(*p2));
			p2->next = p->next;
			p->next = p2;
			p = p2;
		}
	}

	/* Install new host */
	++totalhosts;
	p->addr = a;
	if (fqdn != NULL)
		p->name = strdup(fqdn);
	if (state != 0)
		p->state = state;
	if (p->state == STATE_FREE)
		abort();

	/* Return answer entry */
	return (p);
}

struct htable2 *
cachehost(const char *host, int state)
{
	struct htable2 *p, *p2;

	/* Look for existing slot in hash table */
	for (p = hash2(host); p != NULL; p = p->next)
		if (p->state != STATE_FREE && strcmp(host, p->host) == 0)
			break;

	/* Allocate a new slot */
	if (p == NULL) {
		p = hash2(host);
		if (p->state != STATE_FREE) {
			/* Handle the collision */
			p2 = (struct htable2 *)malloc(sizeof(*p2));
			/* Lose, lose */
			if (p2 == NULL)
				return (NULL);
			memset((char *)p2, 0, sizeof(*p2));
			p2->next = p->next;
			p->next = p2;
			p = p2;
		}
	}

	/* Install new host */
	p->host = strdup(host);
	if (state != 0)
		p->state = state;
	if (p->state == STATE_FREE)
		abort();

	/* Return answer entry */
	return (p);
}

int
compipv4(const void *vp1, const void *vp2)
{
	const char * const *pp1;
	const char * const *pp2;
	u_int32_t a1, a2;

	pp1 = vp1;
	pp2 = vp2;

	memmove(&a1, pp1, NS_INADDRSZ);
	memmove(&a2, pp2, NS_INADDRSZ);

	if (a1 < a2)
		return (-1);
	if (a1 > a2)
		return (1);
	return (0);
}

static inline struct htable *
hash(const u_int32_t a)
{
	u_int32_t h;

	h = (a & 0xff) ^ ((a >> 16) & 0xff);
	return (&htable[h & (HSIZE - 1)]);
}

static inline struct htable2 *
hash2(const char *name)
{
	size_t len;
	u_int32_t h;

	len = strlen(name);
	if (len > sizeof(h))
		len = sizeof(h);
	h = 0;
	memmove(&h, name, len);
	return (&htable2[h & (HSIZE - 1)]);
}

/* Address required, fqdn is optional */
struct htable *
increment(u_int32_t a, const char *fqdn)
{
	const char *cp;
	struct htable *p;
	struct hostent *hp;
	struct in_addr ip;

	if (a == 0)
		return (NULL);
	p = lookupaddr(a);
	if (p == NULL) {
		if (fqdn == NULL) {
			hp = gethostbyaddr(&a, NS_INADDRSZ, AF_INET);
			++addrcnt;
			if (hp == NULL) {
				ip.s_addr = a;
				cp = inet_ntoa(ip);
				if (debug > 1)
					fprintf(stderr,
					    "%s:%u: Can't translate: %s\n",
					    file, lineno, cp);
				fqdn = cp;
			} else {
				lowcase(hp->h_name);
				fqdn = hp->h_name;
				if (hp->h_length != NS_INADDRSZ) {
					fprintf(stderr, "%s:%u: hp->h_length"
					    " != NS_INADDRSZ (%d)\n",
					    file, lineno, hp->h_length);
					return (NULL);
				}

				if (hp->h_addrtype != AF_INET) {
					fprintf(stderr, "%s:%u: hp->h_addrtype"
					    " != AF_INET (%d)\n",
					    file, lineno, hp->h_addrtype);
					return (NULL);
				}
			}
		}
		p = cacheaddr(a, STATE_USED, fqdn);
		if (p == NULL) {
			fprintf(stderr, "increment: cacheaddr failed\n");
			abort();
		}
	}
	++p->count;
	return (p);
}

void
logseek(FILE *f, const char *fn, off_t *offp)
{
	int fd;
	struct stat sbuf;

	if (!doseek)
		return;
	if (doseek && doinode) {
		fd = fileno(f);
		if (fstat(fd, &sbuf) < 0) {
			fprintf(stderr, "%s: fstat: %s\n",
			    fn, strerror(errno));
			return;
		}
		if (inode != sbuf.st_ino) {
			if (debug > 0)
				fprintf(stderr,
				    "%s: inode mismatch (ignoring -O)\n", fn);
			return;
		}
		if (debug > 0)
			fprintf(stderr, "%s: inode matches\n", fn);
		inode = 0;
	}
	/* Only seek if the offset is non-zero (allows bootstrapping) */
	if (*offp > 0) {
		if (fseeko(f, *offp, SEEK_SET) < 0)
			fprintf(stderr, "%s: fseeko: %s\n",
			    fn, strerror(errno));
		*offp = 0;
	}
}

/* Look hash table entry for address */
struct htable *
lookupaddr(u_int32_t a)
{
	struct htable *p;

	for (p = hash(a); p != NULL; p = p->next)
		if (a == p->addr)
			return (p);
	return (NULL);
}

/* Look hash table entry for a simple name */
struct htable2 *
lookuphost(const char *host)
{
	struct htable2 *p;

	for (p = hash2(host); p != NULL; p = p->next)
		if (p->host != NULL && strcmp(host, p->host) == 0)
			return (p);
	return (NULL);
}

void
lowcase(char *str)
{
	char *cp;

	for (cp = str; *cp != '\0'; ++cp)
		if (isupper(*cp))
			*cp = tolower(*cp);
}

/* Important that resolv.conf has a "search" config for all possible domains */
void
name(const char *host, size_t len)
{
	size_t n;
	char **pp;
	struct hostent *hp;
	char thost[256];
	char *fqdn;
	struct htable *p;
	struct htable2 *p2;

	static char lasthost[256] = "?";
	static char lastfqdn[256];
	static struct in_addr lastaddr;
	u_int32_t a;

	(void)snprintf(thost, sizeof(thost), "%.*s", (int)len, host);

	/* Try simple name hash first */
	p2 = lookuphost(thost);
	if (p2 != NULL) {
		++p2->htablep->count;
		return;
	}

	hp = gethostbyname(thost);
	++hostcnt;
	if (hp == NULL) {
		if (debug > 1)
			fprintf(stderr, "%s:%u: gethostbyname(%s) failed\n",
			    file, lineno, thost);
		return;
	}
	lowcase(hp->h_name);
	fqdn = hp->h_name;

	if (hp->h_length != NS_INADDRSZ) {
		fprintf(stderr, "%s:%u: hp->h_length != NS_INADDRSZ (%d)\n",
		    file, lineno, hp->h_length);
		return;
	}

	if (hp->h_addrtype != AF_INET) {
		fprintf(stderr, "%s:%u: hp->h_addrtype != AF_INET (%d)\n",
		    file, lineno, hp->h_addrtype);
		return;
	}


	/* Sort when more than one IPv4 address */
	if (hp->h_addr_list[1] != NULL) {
		for (n = 0, pp = hp->h_addr_list; *pp != NULL; ++n, ++pp)
			continue;
		qsort(*hp->h_addr_list, n, NS_INADDRSZ, compipv4);
	}

	memmove(&a, hp->h_addr_list[0], NS_INADDRSZ);
	p = increment(a, fqdn);
	if (p == NULL) {
		fprintf(stderr, "%s:%u: %s: no addresses found\n",
		    file, lineno, fqdn);
		return;
	}

	p2 = cachehost(thost, STATE_USED);
	p2->htablep = p;

	strcpy(lastfqdn, fqdn);
	strcpy(lasthost, thost);
	memmove(&lastaddr, hp->h_addr_list[0], sizeof(lastaddr));
}

void
summary()
{
	int i, j;
	struct htable *p, *p2;
	char buf[132];
	struct tm *tp;
	time_t now;
	FILE *f;
	int maxcol;
	int emptybkts;

	f = stdout;
	if (timestamp) {
		(void)time(&now);
		tp = localtime(&now);
		strftime(buf, sizeof(buf), "%F %R", tp);
		fprintf(f, "# Updated: %s\n", buf);
	}

	if (stats)
		fprintf(f, "# Total hosts: %d\n", totalhosts);
	if (doinode || doseek) {
		fprintf(f, "# inode %lu\n", (u_long)lastinode);
		fprintf(f, "# offset %lld\n", (long long int)lastoffset);
	}
	if (stats) {
		if (stats > 1) {
			fprintf(f, "# gethostbyaddr() calls: %d\n", addrcnt);
			fprintf(f, "# gethostbyname() calls: %d\n", hostcnt);
			maxcol = 0;
			emptybkts = 0;
			for (p = htable, i = 0; i < HSIZE; ++p, ++i) {
				if (p->state == STATE_FREE) {
					++emptybkts;
					continue;
				}
				j = 0;
				for (p2 = p; p2 != NULL; p2 = p2->next)
					++j;
				if (maxcol < j)
					maxcol = j;
			}
			fprintf(f, "# Empty buckets: %d/%d\n",
			    emptybkts, HSIZE);
			fprintf(f, "# max collisions: %d\n", maxcol);
		}
	}

	for (p = htable, i = 0; i < HSIZE; ++p, ++i) {
		if (p->name == NULL)
			continue;
		for (p2 = p; p2 != NULL; p2 = p2->next) {
			if (inet_ntop(AF_INET, &p2->addr,
			    buf, sizeof(buf)) == NULL)
				strcpy(buf, "?");
			fprintf(f, "%s\t%s\t%d\n", p2->name, buf, p2->count);
		}
	}

}

void
saveinfo(FILE *f)
{
	int fd;
	struct stat sbuf;

	fd = fileno(f);
	if (fstat(fd, &sbuf) < 0) {
		fprintf(stderr, "saveinfo: fstat %s\n", strerror(errno));
		return;
	}
	lastinode = sbuf.st_ino;
	lastoffset = ftello(f);
}

void
usage()
{
	extern char version[];

	(void)fprintf(stderr, "Version %s\n", version);
	(void)fprintf(stderr,
	    "usage: %s [-dts] [-I inode] [-O offset] [file ...]\n", prog);
	exit(1);
}
